#include "gaamscomm.h"

extern volatile int readErrorCount;
extern INFO info;

#ifdef OS_WIN
/**************************************************************/
/** Windows branch                                           **/
/**************************************************************/

static HANDLE rs485_fd;

int rs485_open (const char *fname, const int cbaud, const int databits, const int parity)
{
	rs485_fd = CreateFileA (fname, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (rs485_fd == INVALID_HANDLE_VALUE) return -1;
	if (parity < 0 || parity > 2) return -1;

	DCB port_settings;
	memset (&port_settings, 0, sizeof (port_settings));
	port_settings.DCBlength = sizeof (port_settings);

	if (!GetCommState (rs485_fd, &port_settings)) {
		CloseHandle (rs485_fd);
		return -1;
	}

	port_settings.BaudRate = cbaud;
	port_settings.fBinary = TRUE;
	port_settings.StopBits = ONESTOPBIT;
	port_settings.ByteSize = databits;
	port_settings.fOutxCtsFlow = FALSE;
	port_settings.fOutxDsrFlow = FALSE;
	port_settings.fDtrControl = DTR_CONTROL_DISABLE;
	port_settings.fDsrSensitivity = FALSE;
	port_settings.fTXContinueOnXoff = FALSE;
	port_settings.fRtsControl = RTS_CONTROL_DISABLE;

	if (parity == 0) {
		port_settings.fParity = FALSE;
	}
	if (parity == 1) {
		port_settings.fParity = TRUE;
		port_settings.Parity = EVENPARITY;
	}
	if (parity == 2) {
		port_settings.fParity = TRUE;
		port_settings.Parity = ODDPARITY;
	}

	if (!SetCommState (rs485_fd, &port_settings)) {
		CloseHandle (rs485_fd);
		return -1;
	}

	COMMTIMEOUTS Cptimeouts;

	Cptimeouts.ReadIntervalTimeout         = 100;
	Cptimeouts.ReadTotalTimeoutMultiplier  = 0;
	Cptimeouts.ReadTotalTimeoutConstant    = 150;
	Cptimeouts.WriteTotalTimeoutMultiplier = 100;
	Cptimeouts.WriteTotalTimeoutConstant   = 100;

	if (!SetCommTimeouts (rs485_fd, &Cptimeouts)) {
		CloseHandle (rs485_fd);
		return -1;
	}

	return 0;
}

void rs485_close (void)
{
	CloseHandle (rs485_fd);
}

ssize_t rs485_send (const ssize_t len, const uint8_t *buf)
{
	ssize_t r = 0;

	if (!WriteFile (rs485_fd, buf, len, (LPDWORD) ( (void *) &r), NULL)) {
		r = 0;
	}
	if (r != len) {
		r = 0;
	}
	rs485_debug ("WRITE", r, buf);

	return r;
}

ssize_t rs485_recv (uint8_t *buf)
{
	ssize_t r = 0;
	ReadFile (rs485_fd, buf, BUFSIZE, (LPDWORD) ( (void *) &r), NULL);
	if (r > 0) {
		rs485_debug ("READ", r, buf);
	}
	return r;
}

#else
/**************************************************************/
/** Linux branch                                             **/
/**************************************************************/

int rs485_fd = -1;
struct timeval rs485_read_timeout;
fd_set rs485_read_input;

int rs485_open (const char *fname, const int cbaud, const int databits, const int parity)
{
	struct termios options;

	if (rs485_fd >= 0) return 0;
	rs485_fd = open (fname, O_RDWR | O_NOCTTY | O_SYNC);
	//O_NDELAY
	if (rs485_fd == -1) return -1;
	fcntl (rs485_fd, F_SETFL, 0);

	tcgetattr (rs485_fd, &options);
	cfmakeraw (&options);

	cfsetispeed (&options, cbaud);
	cfsetospeed (&options, cbaud);

	options.c_cflag |= (CLOCAL | CREAD);

	if (parity == 0) { // no parity
		options.c_cflag &= ~PARENB;
	}
	else if (parity == 1) { // even parity
		options.c_cflag |= PARENB;
		options.c_cflag &= ~PARODD;
	}
	else if (parity == 2) { // odd parity
		options.c_cflag |= PARENB;
		options.c_cflag |= PARODD;
	}

	options.c_cflag &= ~CSTOPB;
	options.c_cflag &= ~CSIZE;
	options.c_cflag |= databits;

	options.c_cflag &= ~CRTSCTS;

	tcsetattr (rs485_fd, TCSANOW, &options);

	usleep (500000);
	return 0;
}

void rs485_close (void)
{
	if (rs485_fd >= 0) {
		close (rs485_fd);
	}
	rs485_fd = -1;
}

ssize_t rs485_send (const ssize_t len, const uint8_t *buf)
{
	int status;
	ssize_t r = 0;

	ioctl (rs485_fd, TIOCMGET, &status);
	status |= TIOCM_RTS;
	ioctl (rs485_fd, TIOCMSET, &status);

	r = write (rs485_fd, buf, len);
	if (r != len) {
		r = 0;
	}
	rs485_debug ("WRITE", r, buf);

	ioctl (rs485_fd, TIOCMGET, &status);
	status &= ~TIOCM_RTS;
	tcdrain (rs485_fd);
	ioctl (rs485_fd, TIOCMSET, &status);

	return r;
}

ssize_t rs485_recv (uint8_t *buf)
{
	int bytes, n;
	ssize_t r = 0, d;

	while (1) {
		rs485_read_timeout.tv_sec  = 0;
		rs485_read_timeout.tv_usec = 25000;
		FD_ZERO (&rs485_read_input);
		FD_SET (rs485_fd, &rs485_read_input);
		n = select (rs485_fd + 1, &rs485_read_input, NULL, NULL, &rs485_read_timeout);
		if (n <= 0) break;

		ioctl (rs485_fd, FIONREAD, &bytes);
		if (bytes <= 0) break;
		if (r + bytes > BUFSIZE) {
			r = 0;
			break;
		}

		d = read (rs485_fd, buf + r, bytes);
		if (d != bytes) break;
		r += d;
	}
	if (r > 0) {
		rs485_debug ("READ", r, buf);
	}
	return r;
}
#endif // OS_WIN

void rs485_debug (const char *text, const ssize_t len, const uint8_t *buf)
{
	if (!info.debug) return;
	char tbuf[4096], tbuf2[32];
	int i;
	sprintf (tbuf, "%s %d bytes:", text, (int) len);
	for (i = 0; i < len; i++) {
		sprintf (tbuf2, " %02X", buf[i]);
		strcat (tbuf, tbuf2);
	}
	p_printf ("%s", tbuf);
}
