#include "gaamscomm.h"

extern INFO info;

CURLcode _tcpclient_wait_on_socket (GAAMSCOMM_CURL *c, const int for_recv, const long timeout_ms)
{
	struct timeval tv;
	fd_set infd, outfd, errfd;

	tv.tv_sec = timeout_ms / 1000;
	tv.tv_usec = (timeout_ms % 1000) * 1000;

	FD_ZERO (&infd);
	FD_ZERO (&outfd);
	FD_ZERO (&errfd);

	FD_SET (c->sockfd, &errfd); /* always check for error */

	if (for_recv) {
		FD_SET (c->sockfd, &infd);
	} else {
		FD_SET (c->sockfd, &outfd);
	}

	/* select() returns the number of signalled sockets or -1 */
	int res = select (c->sockfd + 1, &infd, &outfd, &errfd, &tv);

	if (res < 0) {
		return CURLE_COULDNT_CONNECT;
	}

	return (res > 0) ? CURLE_OK : CURLE_COULDNT_CONNECT;
}

CURLcode _tcpclient_write (GAAMSCOMM_CURL *c, const char *buf)
{
	ssize_t iolen;
	size_t total = strlen (buf);
	size_t pos = 0;

    c->sockfd = socket (AF_INET, SOCK_STREAM, 0);

#ifdef OS_WIN
	if (c->sockfd == INVALID_SOCKET) {
		return CURLE_COULDNT_CONNECT;
	}

	unsigned long mode = 1;
	if (ioctlsocket (c->sockfd, FIONBIO, &mode) != 0) {
		return CURLE_COULDNT_CONNECT;
	}

	if (connect (c->sockfd, (SOCKADDR *) &(info.serveraddr), sizeof(info.serveraddr)) != 0) {
		const int err = WSAGetLastError ();
		if (err != WSAEWOULDBLOCK && err != WSAEALREADY) {
			return CURLE_COULDNT_CONNECT;
		}
	}
#else
    if (c->sockfd < 0) {
		return CURLE_COULDNT_CONNECT;
    }

	int flags = fcntl (c->sockfd, F_GETFL, 0);
	if (flags < 0) {
		return CURLE_COULDNT_CONNECT;
	}
	flags |= O_NONBLOCK;
	if (fcntl (c->sockfd, F_SETFL, flags) != 0) {
		return CURLE_COULDNT_CONNECT;
	}

	if (connect (c->sockfd, (struct sockaddr *) &(info.serveraddr), sizeof(info.serveraddr)) != 0) {
		if (errno != EINPROGRESS) {
			return CURLE_COULDNT_CONNECT;
		}
	}
#endif

	while (pos != total) {
		if (_tcpclient_wait_on_socket (c, 0, 1500L) != CURLE_OK) {
			return CURLE_COULDNT_CONNECT;
		}

		iolen = send (c->sockfd, buf + pos, total - pos, 0);
#ifdef OS_WIN
		if (iolen == SOCKET_ERROR) {
			return CURLE_SEND_ERROR;
		}
#else
		if (iolen < 0) {
			return CURLE_SEND_ERROR;
		}
		else if (iolen == 0) {
			if (errno != EAGAIN && errno != EWOULDBLOCK) {
				return CURLE_SEND_ERROR;
			}
		}
#endif // OS_WIN

		pos += iolen;
		if (pos > total) {
			return CURLE_SEND_ERROR;
		}
	}

#ifdef OS_WIN
	shutdown (c->sockfd, SD_SEND);
#else
	shutdown (c->sockfd, SHUT_WR);
#endif
	return CURLE_OK;
}

int8_t tcpclient_send (GAAMSCOMM_CURL *c, const GAAMSCOMM_EVENT *ev)
{
	int32_t ret;
	char buf[4096];

	if ( (ret = gaamscomm_prepare_tcp_output (ev, buf, sizeof (buf))) != 0) {
		return ret;
	}
	//p_printf ("TCP: %s", buf);

	if (_tcpclient_write (c, buf) != CURLE_OK) {
#ifdef OS_WIN
		if (c->sockfd != INVALID_SOCKET) {
			closesocket (c->sockfd);
		}
#else
		if (c->sockfd >= 0) {
			close (c->sockfd);
		}
#endif
		return -5;
	}

#ifdef OS_WIN
	closesocket (c->sockfd);
#else
	close (c->sockfd);
#endif

	return 1;
}

int8_t tcpclient_global_open (void)
{
	if (info.url_host == NULL || info.url_port == NULL) {
		return -1;
	}

#ifdef OS_WIN
	WSADATA wsaData;
	if (WSAStartup (MAKEWORD(2,2), &wsaData) != 0) {
		WSACleanup ();
		return -1;
	}
#endif

    struct addrinfo hints;
    struct addrinfo *result, *rp;

	memset (&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;

	char host[NI_MAXHOST] = "";
	int ret;

	ret = getaddrinfo (info.url_host, info.url_port, &hints, &result);
	if (ret != 0) {
		p_printf ("DNS error: %s", gai_strerror(ret));
		return -5;
	}

	for (rp = result; rp != NULL; rp = rp->ai_next) {
		if (rp->ai_family == AF_INET) {
			memcpy (&(info.serveraddr), (struct sockaddr_in*)rp->ai_addr, sizeof (struct sockaddr_in));
			ret = getnameinfo (rp->ai_addr, rp->ai_addrlen, host, sizeof (host), NULL, 0, NI_NUMERICHOST);
			if (ret != 0) {
				p_printf ("getnameinfo() failed: %s", gai_strerror (ret));
			}
			p_printf ("DNS translation: %s -> %s", info.url_host, host);
			break;
		}
	}

	if (rp == NULL) {
		freeaddrinfo(result);
		return -6;
	}

	freeaddrinfo(result);

	return 0;
}

void tcpclient_global_close (void)
{
#ifdef OS_WIN
	WSACleanup ();
#endif
}
